package sun.rmi.transport.tcp;

import java.io.OutputStream;
import java.io.InputStream;
import java.io.BufferedOutputStream;
import sun.rmi.transport.proxy.HttpReceiveSocket;
import java.io.BufferedInputStream;
import java.lang.ref.SoftReference;
import java.rmi.server.RMIFailureHandler;
import java.lang.reflect.InvocationTargetException;
import java.net.InetAddress;
import java.lang.reflect.UndeclaredThrowableException;
import java.util.logging.Level;
import java.util.concurrent.RejectedExecutionException;
import java.security.Permissions;
import java.security.ProtectionDomain;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.TimeUnit;
import sun.security.action.GetLongAction;
import sun.security.action.GetIntegerAction;
import java.rmi.server.LogStream;
import sun.security.action.GetPropertyAction;
import sun.rmi.transport.Channel;
import java.rmi.server.ServerNotActiveException;
import sun.rmi.transport.DGCAckHandler;
import java.rmi.server.UID;
import java.io.DataOutputStream;
import sun.rmi.transport.StreamRemoteCall;
import java.io.DataInputStream;
import sun.rmi.transport.Connection;
import java.net.Socket;
import java.net.BindException;
import java.rmi.server.ExportException;
import java.security.AccessController;
import sun.rmi.runtime.NewThreadAction;
import java.io.IOException;
import java.rmi.RemoteException;
import sun.rmi.transport.Target;
import java.lang.ref.WeakReference;
import sun.rmi.transport.Endpoint;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.WeakHashMap;
import java.rmi.server.RMISocketFactory;
import java.lang.ref.Reference;
import java.util.Map;
import java.net.ServerSocket;
import java.util.LinkedList;
import java.security.AccessControlContext;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.ExecutorService;
import sun.rmi.runtime.Log;
import sun.rmi.transport.Transport;

public class TCPTransport extends Transport
{
    static final Log tcpLog;
    private static final int maxConnectionThreads;
    private static final long threadKeepAliveTime;
    private static final ExecutorService connectionThreadPool;
    private static final boolean disableIncomingHttp;
    private static final AtomicInteger connectionCount;
    private static final ThreadLocal<TCPTransport.ConnectionHandler> threadConnectionHandler;
    private static final AccessControlContext NOPERMS_ACC;
    private final LinkedList<TCPEndpoint> epList;
    private int exportCount;
    private ServerSocket server;
    private final Map<TCPEndpoint, Reference<TCPChannel>> channelTable;
    static final RMISocketFactory defaultSocketFactory;
    private static final int connectionReadTimeout;
    
    TCPTransport(LinkedList<TCPEndpoint> epList) {
        this.exportCount = 0;
        this.server = null;
        this.channelTable = new WeakHashMap();
        this.epList = epList;
        if (TCPTransport.tcpLog.isLoggable(Log.BRIEF)) {
            TCPTransport.tcpLog.log(Log.BRIEF, "Version = 2, ep = " + this.getEndpoint());
        }
    }
    
    public void shedConnectionCaches() {
        ArrayList list;
        synchronized (this.channelTable) {
            list = new ArrayList(this.channelTable.values().size());
            Iterator<Reference<TCPChannel>> iterator = this.channelTable.values().iterator();
            while (iterator.hasNext()) {
                TCPChannel tcpChannel = (TCPChannel)((Reference)iterator.next()).get();
                if (tcpChannel != null) {
                    list.add(tcpChannel);
                }
                else {
                    continue;
                }
            }
        }
        Iterator<Object> iterator2 = list.iterator();
        while (iterator2.hasNext()) {
            ((TCPChannel)iterator2.next()).shedCache();
        }
    }
    
    @Override
    public TCPChannel getChannel(Endpoint endpoint) {
        Object referent = null;
        if ((endpoint instanceof TCPEndpoint)) {
            synchronized (this.channelTable) {
                Reference<Object> reference = (Reference)this.channelTable.get(endpoint);
                if (reference != null) {
                    referent = (Object)reference.get();
                }
                if (referent == null) {
                    TCPEndpoint tcpEndpoint = (TCPEndpoint)endpoint;
                    referent = new TCPChannel(this, tcpEndpoint);
                    this.channelTable.put(tcpEndpoint, new WeakReference(referent));
                }
            }
        }
        return referent;
    }
    
    @Override
    public void free(Endpoint endpoint) {
        if ((endpoint instanceof TCPEndpoint)) {
            synchronized (this.channelTable) {
                Reference<Object> reference = (Reference)this.channelTable.remove(endpoint);
                if (reference != null) {
                    TCPChannel tcpChannel = (TCPChannel)reference.get();
                    if (tcpChannel != null) {
                        tcpChannel.shedCache();
                    }
                }
            }
        }
    }
    
    @Override
    public void exportObject(Target target) throws RemoteException {
        synchronized (this) {
            this.listen();
            ++this.exportCount;
        }
        boolean n = false;
        try {
            super.exportObject(target);
            n = true;
        }
        finally {
            if (n == 0) {
                synchronized (this) {
                    this.decrementExportCount();
                }
            }
        }
    }
    
    @Override
    protected synchronized void targetUnexported() {
        this.decrementExportCount();
    }
    
    private void decrementExportCount() {
        assert !(!Thread.holdsLock(this));
        --this.exportCount;
        if (this.exportCount == 0 && this.getEndpoint().getListenPort() != 0) {
            ServerSocket server = this.server;
            this.server = null;
            try {
                server.close();
            }
            catch (IOException ex) {}
        }
        return;
    }
    
    @Override
    protected void checkAcceptPermission(AccessControlContext accessControlContext) {
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager == null) {
            return;
        }
        TCPTransport.ConnectionHandler connectionHandler = (TCPTransport.ConnectionHandler)TCPTransport.threadConnectionHandler.get();
        if (connectionHandler == null) {
            throw new Error("checkAcceptPermission not in ConnectionHandler thread");
        }
        connectionHandler.checkAcceptPermission(securityManager, accessControlContext);
        return;
    }
    
    private TCPEndpoint getEndpoint() {
        synchronized (this.epList) {
            return (TCPEndpoint)this.epList.getLast();
        }
    }
    
    private void listen() throws RemoteException {
        assert !(!Thread.holdsLock(this));
        TCPEndpoint endpoint = this.getEndpoint();
        int port = endpoint.getPort();
        if (this.server == null) {
            if (TCPTransport.tcpLog.isLoggable(Log.BRIEF)) {
                TCPTransport.tcpLog.log(Log.BRIEF, "(port " + port + ") create server socket");
            }
            try {
                this.server = endpoint.newServerSocket();
                ((Thread)AccessController.doPrivileged(new NewThreadAction(new TCPTransport.AcceptLoop(this.server), "TCP Accept-" + port, true))).start();
            }
            catch (BindException ex) {
                throw new ExportException("Port already in use: " + port, ex);
            }
            catch (IOException ex2) {
                throw new ExportException("Listen failed on port: " + port, ex2);
            }
        }
        else {
            SecurityManager securityManager = System.getSecurityManager();
            if (securityManager != null) {
                securityManager.checkListen(port);
            }
        }
        return;
    }
    
    private static void closeSocket(Socket socket) {
        try {
            socket.close();
        }
        catch (IOException ex) {}
    }
    
    void handleMessages(Connection c, boolean b) {
        int port = this.getEndpoint().getPort();
        try {
            DataInputStream in = new DataInputStream(c.getInputStream());
            do {
                int read = in.read();
                if (read == -1) {
                    if (TCPTransport.tcpLog.isLoggable(Log.BRIEF)) {
                        TCPTransport.tcpLog.log(Log.BRIEF, "(port " + port + ") connection closed");
                        break;
                    }
                    break;
                }
                else {
                    if (TCPTransport.tcpLog.isLoggable(Log.BRIEF)) {
                        TCPTransport.tcpLog.log(Log.BRIEF, "(port " + port + ") op = " + read);
                    }
                    switch (read) {
                        case 80: {
                            if (!this.serviceCall(new StreamRemoteCall(c))) {
                                return;
                            }
                            continue;
                        }
                        case 82: {
                            new DataOutputStream(c.getOutputStream()).writeByte(83);
                            c.releaseOutputStream();
                            continue;
                        }
                        case 84: {
                            DGCAckHandler.received(UID.read(in));
                            continue;
                        }
                        default: {
                            throw new IOException("unknown transport op " + read);
                        }
                    }
                }
            } while (b);
        }
        catch (IOException ex) {
            if (TCPTransport.tcpLog.isLoggable(Log.BRIEF)) {
                TCPTransport.tcpLog.log(Log.BRIEF, "(port " + port + ") exception: ", ex);
            }
        }
        finally {
            try {
                c.close();
            }
            catch (IOException ex2) {}
        }
    }
    
    public static String getClientHost() throws ServerNotActiveException {
        TCPTransport.ConnectionHandler connectionHandler = (TCPTransport.ConnectionHandler)TCPTransport.threadConnectionHandler.get();
        if (connectionHandler != null) {
            return connectionHandler.getClientHost();
        }
        throw new ServerNotActiveException("not in a remote call");
    }
    
    static TCPEndpoint access$000(TCPTransport tcpTransport) {
        return tcpTransport.getEndpoint();
    }
    
    static void access$200(Socket socket) {
        closeSocket(socket);
    }
    
    static {
        tcpLog = Log.getLog("sun.rmi.transport.tcp", "tcp", LogStream.parseLevel((String)AccessController.doPrivileged(new GetPropertyAction("sun.rmi.transport.tcp.logLevel"))));
        maxConnectionThreads = (int)AccessController.doPrivileged(new GetIntegerAction("sun.rmi.transport.tcp.maxConnectionThreads", Integer.MAX_VALUE));
        threadKeepAliveTime = (long)AccessController.doPrivileged(new GetLongAction("sun.rmi.transport.tcp.threadKeepAliveTime", 60000L));
        connectionThreadPool = new ThreadPoolExecutor(0, TCPTransport.maxConnectionThreads, TCPTransport.threadKeepAliveTime, TimeUnit.MILLISECONDS, new SynchronousQueue(), new ThreadFactory() {
            @Override
            public Thread newThread(Runnable runnable) {
                return (Thread)AccessController.doPrivileged(new NewThreadAction(runnable, "TCP Connection(idle)", true, true));
            }
        });
        disableIncomingHttp = ((String)AccessController.doPrivileged(new GetPropertyAction("sun.rmi.server.disableIncomingHttp", "true"))).equalsIgnoreCase("true");
        connectionCount = new AtomicInteger(0);
        threadConnectionHandler = new ThreadLocal();
        NOPERMS_ACC = new AccessControlContext(new ProtectionDomain[] { new ProtectionDomain(null, new Permissions()) });
        defaultSocketFactory = RMISocketFactory.getDefaultSocketFactory();
        connectionReadTimeout = (int)AccessController.doPrivileged(new GetIntegerAction("sun.rmi.transport.tcp.readTimeout", 7200000));
    }
    
    private class AcceptLoop implements Runnable
    {
        private final ServerSocket serverSocket;
        private long lastExceptionTime;
        private int recentExceptionCount;
        
        AcceptLoop(ServerSocket serverSocket) {
            this.lastExceptionTime = 0L;
            this.serverSocket = serverSocket;
        }
        
        @Override
        public void run() {
            try {
                this.executeAcceptLoop();
            }
            finally {
                try {
                    this.serverSocket.close();
                }
                catch (IOException ex) {}
            }
        }
        
        private void executeAcceptLoop() {
            if (TCPTransport.tcpLog.isLoggable(Log.BRIEF)) {
                TCPTransport.tcpLog.log(Log.BRIEF, "listening on port " + TCPTransport.access$000(TCPTransport.this).getPort());
            }
            while (true) {
                Socket accept = null;
                try {
                    accept = this.serverSocket.accept();
                    InetAddress inetAddress = accept.getInetAddress();
                    String str = (inetAddress != null) ? inetAddress.getHostAddress() : "0.0.0.0";
                    try {
                        TCPTransport.connectionThreadPool.execute(new TCPTransport.ConnectionHandler(accept, str));
                    }
                    catch (RejectedExecutionException ex) {
                        TCPTransport.access$200(accept);
                        TCPTransport.tcpLog.log(Log.BRIEF, "rejected connection from " + str);
                    }
                }
                catch (Throwable undeclaredThrowable) {
                    try {
                        if (this.serverSocket.isClosed()) {
                            break;
                        }
                        try {
                            if (TCPTransport.tcpLog.isLoggable(Level.WARNING)) {
                                TCPTransport.tcpLog.log(Level.WARNING, "accept loop for " + this.serverSocket + " throws", undeclaredThrowable);
                            }
                        }
                        catch (Throwable t) {}
                    }
                    finally {
                        if (accept != null) {
                            TCPTransport.access$200(accept);
                        }
                    }
                    if (!(undeclaredThrowable instanceof SecurityException)) {
                        try {
                            TCPEndpoint.shedConnectionCaches();
                        }
                        catch (Throwable t2) {}
                    }
                    if (undeclaredThrowable instanceof Exception || undeclaredThrowable instanceof OutOfMemoryError || undeclaredThrowable instanceof NoClassDefFoundError) {
                        if (!this.continueAfterAcceptFailure(undeclaredThrowable)) {
                            return;
                        }
                        continue;
                    }
                    else {
                        if ((undeclaredThrowable instanceof Error)) {
                            throw (Error)undeclaredThrowable;
                        }
                        throw new UndeclaredThrowableException(undeclaredThrowable);
                    }
                }
            }
        }
        
        private boolean continueAfterAcceptFailure(Throwable target) {
            RMIFailureHandler failureHandler = RMISocketFactory.getFailureHandler();
            if (failureHandler != null) {
                return failureHandler.failure((target instanceof Exception) ? ((Exception)target) : new InvocationTargetException(target));
            }
            this.throttleLoopOnException();
            return true;
        }
        
        private void throttleLoopOnException() {
            long currentTimeMillis = System.currentTimeMillis();
            if (this.lastExceptionTime == 0L || currentTimeMillis - this.lastExceptionTime > 5000L) {
                this.lastExceptionTime = currentTimeMillis;
                this.recentExceptionCount = 0;
            }
            else if (++this.recentExceptionCount >= 10) {
                try {
                    Thread.sleep(10000L);
                }
                catch (InterruptedException ex) {}
            }
        }
    }
    
    private class ConnectionHandler implements Runnable
    {
        private static final int POST = 1347375956;
        private AccessControlContext okContext;
        private Map<AccessControlContext, Reference<AccessControlContext>> authCache;
        private SecurityManager cacheSecurityManager;
        private Socket socket;
        private String remoteHost;
        
        ConnectionHandler(Socket socket, String remoteHost) {
            this.cacheSecurityManager = null;
            this.socket = socket;
            this.remoteHost = remoteHost;
        }
        
        String getClientHost() {
            return this.remoteHost;
        }
        
        void checkAcceptPermission(SecurityManager cacheSecurityManager, AccessControlContext accessControlContext) {
            if (cacheSecurityManager != this.cacheSecurityManager) {
                this.okContext = null;
                this.authCache = new WeakHashMap();
                this.cacheSecurityManager = cacheSecurityManager;
            }
            if (accessControlContext.equals(this.okContext) || this.authCache.containsKey(accessControlContext)) {
                return;
            }
            InetAddress inetAddress = this.socket.getInetAddress();
            cacheSecurityManager.checkAccept((inetAddress != null) ? inetAddress.getHostAddress() : "*", this.socket.getPort());
            this.authCache.put(accessControlContext, new SoftReference(accessControlContext));
            this.okContext = accessControlContext;
            return;
        }
        
        @Override
        public void run() {
            Thread currentThread = Thread.currentThread();
            String name = currentThread.getName();
            try {
                currentThread.setName("RMI TCP Connection(" + TCPTransport.connectionCount.incrementAndGet() + ")-" + this.remoteHost);
                AccessController.doPrivileged(()/*  */ -> {
                    this.run0();
                    return null;
                }, TCPTransport.NOPERMS_ACC);
            }
            finally {
                currentThread.setName(name);
            }
        }
        
        private void run0() {
            TCPEndpoint access$000 = TCPTransport.access$000(TCPTransport.this);
            int port = access$000.getPort();
            TCPTransport.threadConnectionHandler.set(this);
            try {
                this.socket.setTcpNoDelay(true);
            }
            catch (Exception ex) {}
            try {
                if (TCPTransport.connectionReadTimeout > 0) {
                    this.socket.setSoTimeout(TCPTransport.connectionReadTimeout);
                }
            }
            catch (Exception ex2) {}
            try {
                InputStream inputStream = this.socket.getInputStream();
                InputStream inputStream2 = inputStream.markSupported() ? inputStream : new BufferedInputStream(inputStream);
                inputStream2.mark(4);
                DataInputStream o = new DataInputStream(inputStream2);
                int n = o.readInt();
                if (n == 1347375956) {
                    if (TCPTransport.disableIncomingHttp) {
                        throw new RemoteException("RMI over HTTP is disabled");
                    }
                    TCPTransport.tcpLog.log(Log.BRIEF, "decoding HTTP-wrapped call");
                    inputStream2.reset();
                    try {
                        this.socket = new HttpReceiveSocket(this.socket, inputStream2, null);
                        this.remoteHost = "0.0.0.0";
                        inputStream2 = new BufferedInputStream(this.socket.getInputStream());
                        o = new DataInputStream(inputStream2);
                        n = o.readInt();
                    }
                    catch (IOException cause) {
                        throw new RemoteException("Error HTTP-unwrapping call", cause);
                    }
                }
                int short1 = o.readShort();
                if (n != 1246907721 || short1 != 2) {
                    TCPTransport.access$200(this.socket);
                    return;
                }
                OutputStream outputStream = this.socket.getOutputStream();
                BufferedOutputStream out = new BufferedOutputStream(outputStream);
                DataOutputStream o2 = new DataOutputStream(out);
                int port2 = this.socket.getPort();
                if (TCPTransport.tcpLog.isLoggable(Log.BRIEF)) {
                    TCPTransport.tcpLog.log(Log.BRIEF, "accepted socket from [" + this.remoteHost + ":" + port2 + "]");
                }
                switch (o.readByte()) {
                    case 76: {
                        TCPTransport.this.handleMessages(new TCPConnection(new TCPChannel(TCPTransport.this, new TCPEndpoint(this.remoteHost, this.socket.getLocalPort(), access$000.getClientSocketFactory(), access$000.getServerSocketFactory())), this.socket, inputStream2, out), false);
                        break;
                    }
                    case 75: {
                        o2.writeByte(78);
                        if (TCPTransport.tcpLog.isLoggable(Log.VERBOSE)) {
                            TCPTransport.tcpLog.log(Log.VERBOSE, "(port " + port + ") suggesting " + this.remoteHost + ":" + port2);
                        }
                        o2.writeUTF(this.remoteHost);
                        o2.writeInt(port2);
                        o2.flush();
                        String utf = o.readUTF();
                        int int1 = o.readInt();
                        if (TCPTransport.tcpLog.isLoggable(Log.VERBOSE)) {
                            TCPTransport.tcpLog.log(Log.VERBOSE, "(port " + port + ") client using " + utf + ":" + int1);
                        }
                        TCPTransport.this.handleMessages(new TCPConnection(new TCPChannel(TCPTransport.this, new TCPEndpoint(this.remoteHost, this.socket.getLocalPort(), access$000.getClientSocketFactory(), access$000.getServerSocketFactory())), this.socket, inputStream2, out), true);
                        break;
                    }
                    case 77: {
                        if (TCPTransport.tcpLog.isLoggable(Log.VERBOSE)) {
                            TCPTransport.tcpLog.log(Log.VERBOSE, "(port " + port + ") accepting multiplex protocol");
                        }
                        o2.writeByte(78);
                        if (TCPTransport.tcpLog.isLoggable(Log.VERBOSE)) {
                            TCPTransport.tcpLog.log(Log.VERBOSE, "(port " + port + ") suggesting " + this.remoteHost + ":" + port2);
                        }
                        o2.writeUTF(this.remoteHost);
                        o2.writeInt(port2);
                        o2.flush();
                        TCPEndpoint o3 = new TCPEndpoint(o.readUTF(), o.readInt(), access$000.getClientSocketFactory(), access$000.getServerSocketFactory());
                        if (TCPTransport.tcpLog.isLoggable(Log.VERBOSE)) {
                            TCPTransport.tcpLog.log(Log.VERBOSE, "(port " + port + ") client using " + o3.getHost() + ":" + o3.getPort());
                        }
                        Object o4;
                        synchronized (TCPTransport.this.channelTable) {
                            TCPChannel channel = TCPTransport.this.getChannel(o3);
                            o4 = new ConnectionMultiplexer(channel, inputStream2, outputStream, false);
                            channel.useMultiplexer(o4);
                        }
                        o4.run();
                        break;
                    }
                    default: {
                        o2.writeByte(79);
                        o2.flush();
                        break;
                    }
                }
            }
            catch (IOException ex3) {
                TCPTransport.tcpLog.log(Log.BRIEF, "terminated with exception:", ex3);
            }
            finally {
                TCPTransport.access$200(this.socket);
            }
        }
    }
}
